@@ -23,20 +23,16 @@
  * License for more details.
  */
 
-#include <Python.h>
-#include <structmember.h>
-#include <string.h>
-
 #define PSYCOPG_MODULE
-#include "psycopg/config.h"
-#include "psycopg/python.h"
 #include "psycopg/psycopg.h"
+
 #include "psycopg/lobject.h"
 #include "psycopg/connection.h"
 #include "psycopg/microprotocols.h"
 #include "psycopg/microprotocols_proto.h"
 #include "psycopg/pqpath.h"
-#include "pgversion.h"
+
+#include <string.h>
 
 
 #ifdef PSYCOPG_EXTENSIONS
@@ -51,13 +47,11 @@
 static PyObject *
 psyco_lobj_close(lobjectObject *self, PyObject *args)
 {
-    if (!PyArg_ParseTuple(args, "")) return NULL;
-
     /* file-like objects can be closed multiple times and remember that
        closing the current transaction is equivalent to close all the
        opened large objects */
     if (!lobject_is_closed(self)
-        && self->conn->isolation_level > 0
+        && !self->conn->autocommit
 	&& self->conn->mark == self->mark)
     {
         Dprintf("psyco_lobj_close: closing lobject at %p", self);
@@ -77,18 +71,48 @@ psyco_lobj_close(lobjectObject *self, PyObject *args)
 static PyObject *
 psyco_lobj_write(lobjectObject *self, PyObject *args)
 {
-    int len, res=0;
-    const char *buffer;
+    char *buffer;
+    Py_ssize_t len;
+    Py_ssize_t res;
+    PyObject *obj;
+    PyObject *data = NULL;
+    PyObject *rv = NULL;
 
-    if (!PyArg_ParseTuple(args, "s#", &buffer, &len)) return NULL;
+    if (!PyArg_ParseTuple(args, "O", &obj)) return NULL;
 
     EXC_IF_LOBJ_CLOSED(self);
     EXC_IF_LOBJ_LEVEL0(self);
     EXC_IF_LOBJ_UNMARKED(self);
 
-    if ((res = lobject_write(self, buffer, len)) < 0) return NULL;
+    if (Bytes_Check(obj)) {
+        Py_INCREF(obj);
+        data = obj;
+    }
+    else if (PyUnicode_Check(obj)) {
+        if (!(data = PyUnicode_AsEncodedString(obj, self->conn->codec, NULL))) {
+            goto exit;
+        }
+    }
+    else {
+        PyErr_Format(PyExc_TypeError,
+            "lobject.write requires a string; got %s instead",
+            Py_TYPE(obj)->tp_name);
+        goto exit;
+    }
 
-    return PyInt_FromLong((long)res);
+    if (-1 == Bytes_AsStringAndSize(data, &buffer, &len)) {
+        goto exit;
+    }
+
+    if (0 > (res = lobject_write(self, buffer, (size_t)len))) {
+        goto exit;
+    }
+
+    rv = PyInt_FromLong((long)res);
+
+exit:
+    Py_XDECREF(data);
+    return rv;
 }
 
 /* read method - read data from the lobject */
@@ -100,10 +124,11 @@ static PyObject *
 psyco_lobj_read(lobjectObject *self, PyObject *args)
 {
     PyObject *res;
-    int where, end, size = -1;
+    int where, end;
+    Py_ssize_t size = -1;
     char *buffer;
 
-    if (!PyArg_ParseTuple(args, "|i", &size)) return NULL;
+    if (!PyArg_ParseTuple(args, "|" CONV_CODE_PY_SSIZE_T,  &size)) return NULL;
 
     EXC_IF_LOBJ_CLOSED(self);
     EXC_IF_LOBJ_LEVEL0(self);
@@ -125,9 +150,13 @@ psyco_lobj_read(lobjectObject *self, PyObject *args)
         return NULL;
     }
 
-    res = PyString_FromStringAndSize(buffer, size);
+    if (self->mode & LOBJECT_BINARY) {
+        res = Bytes_FromStringAndSize(buffer, size);
+    } else {
+        res = PyUnicode_Decode(buffer, size, self->conn->codec, NULL);
+    }
     PyMem_Free(buffer);
-    
+
     return res;
 }
 
@@ -165,8 +194,6 @@ psyco_lobj_tell(lobjectObject *self, PyObject *args)
 {
     int pos;
 
-    if (!PyArg_ParseTuple(args, "")) return NULL;
-
     EXC_IF_LOBJ_CLOSED(self);
     EXC_IF_LOBJ_LEVEL0(self);
     EXC_IF_LOBJ_UNMARKED(self);
@@ -185,8 +212,6 @@ psyco_lobj_tell(lobjectObject *self, PyObject *args)
 static PyObject *
 psyco_lobj_unlink(lobjectObject *self, PyObject *args)
 {
-    if (!PyArg_ParseTuple(args, "")) return NULL;
-
     if (lobject_unlink(self) < 0)
     	return NULL;
 
@@ -266,11 +291,11 @@ static struct PyMethodDef lobjectObject_methods[] = {
     {"seek", (PyCFunction)psyco_lobj_seek,
      METH_VARARGS, psyco_lobj_seek_doc},
     {"tell", (PyCFunction)psyco_lobj_tell,
-     METH_VARARGS, psyco_lobj_tell_doc},
+     METH_NOARGS, psyco_lobj_tell_doc},
     {"close", (PyCFunction)psyco_lobj_close,
-     METH_VARARGS, psyco_lobj_close_doc},
+     METH_NOARGS, psyco_lobj_close_doc},
     {"unlink",(PyCFunction)psyco_lobj_unlink,
-     METH_VARARGS, psyco_lobj_unlink_doc},
+     METH_NOARGS, psyco_lobj_unlink_doc},
     {"export",(PyCFunction)psyco_lobj_export,
      METH_VARARGS, psyco_lobj_export_doc},
 #if PG_VERSION_HEX >= 0x080300
@@ -283,10 +308,10 @@ static struct PyMethodDef lobjectObject_methods[] = {
 /* object member list */
 
 static struct PyMemberDef lobjectObject_members[] = {
-    {"oid", T_UINT, offsetof(lobjectObject, oid), RO,
+    {"oid", T_UINT, offsetof(lobjectObject, oid), READONLY,
         "The backend OID associated to this lobject."},
-    {"mode", T_STRING, offsetof(lobjectObject, smode), RO,
-        "Open mode ('r', 'w', 'rw' or 'n')."},
+    {"mode", T_STRING, offsetof(lobjectObject, smode), READONLY,
+        "Open mode."},
     {NULL}
 };
 
@@ -302,12 +327,12 @@ static struct PyGetSetDef lobjectObject_getsets[] = {
 
 static int
 lobject_setup(lobjectObject *self, connectionObject *conn,
-              Oid oid, int mode, Oid new_oid, const char *new_file)
+              Oid oid, const char *smode, Oid new_oid, const char *new_file)
 {
     Dprintf("lobject_setup: init lobject object at %p", self);
 
-    if (conn->isolation_level == 0) {
-        psyco_set_error(ProgrammingError, (PyObject*)self,
+    if (conn->autocommit) {
+        psyco_set_error(ProgrammingError, NULL,
             "can't use a lobject outside of transactions", NULL, NULL);
         return -1;
     }
@@ -320,11 +345,11 @@ lobject_setup(lobjectObject *self, connectionObject *conn,
     self->fd = -1;
     self->oid = InvalidOid;
 
-    if (lobject_open(self, conn, oid, mode, new_oid, new_file) == -1)
+    if (0 != lobject_open(self, conn, oid, smode, new_oid, new_file))
         return -1;
 
    Dprintf("lobject_setup: good lobject object at %p, refcnt = "
-           FORMAT_CODE_PY_SSIZE_T, self, ((PyObject *)self)->ob_refcnt);
+           FORMAT_CODE_PY_SSIZE_T, self, Py_REFCNT(self));
    Dprintf("lobject_setup:    oid = %d, fd = %d", self->oid, self->fd);
    return 0;
 }
@@ -337,27 +362,28 @@ lobject_dealloc(PyObject* obj)
     if (lobject_close(self) < 0)
         PyErr_Print();
     Py_XDECREF((PyObject*)self->conn);
+    PyMem_Free(self->smode);
 
     Dprintf("lobject_dealloc: deleted lobject object at %p, refcnt = "
-            FORMAT_CODE_PY_SSIZE_T, obj, obj->ob_refcnt);
+            FORMAT_CODE_PY_SSIZE_T, obj, Py_REFCNT(obj));
 
-    obj->ob_type->tp_free(obj);
+    Py_TYPE(obj)->tp_free(obj);
 }
 
 static int
 lobject_init(PyObject *obj, PyObject *args, PyObject *kwds)
 {
     Oid oid=InvalidOid, new_oid=InvalidOid;
-    int mode=0;
+    const char *smode = "";
     const char *new_file = NULL;
     PyObject *conn;
 
-    if (!PyArg_ParseTuple(args, "O|iiis",
-         &conn, &oid, &mode, &new_oid, &new_file))
+    if (!PyArg_ParseTuple(args, "O|iziz",
+         &conn, &oid, &smode, &new_oid, &new_file))
         return -1;
 
     return lobject_setup((lobjectObject *)obj,
-        (connectionObject *)conn, oid, mode, new_oid, new_file);
+        (connectionObject *)conn, oid, smode, new_oid, new_file);
 }
 
 static PyObject *
@@ -386,8 +412,7 @@ lobject_repr(lobjectObject *self)
 "A database large object."
 
 PyTypeObject lobjectType = {
-    PyObject_HEAD_INIT(NULL)
-    0,
+    PyVarObject_HEAD_INIT(NULL, 0)
     "psycopg2._psycopg.lobject",
     sizeof(lobjectObject),
     0,
